#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          25-06-2019
//      Created for Project :   FLAME
"""


# Python standard libs
import logging

# PIP installed libs
from influxdb import InfluxDBClient
from pyramid.view import view_defaults, view_config
from requests import delete

# initialise logger
log = logging.getLogger('service_logger')


@view_defaults(renderer='json')
class DataManagementAPI(object):
    """
    A class-based view for managing SFC data - time-series measurements and alerts.
    """

    def __init__(self, request):
        """
        Initialises the instance of the view with the request argument.

        :param request: client's call request
        """

        self.request = request

    @view_config(route_name='data_management_aggregate', request_method='DELETE')
    def delete_all_sfc_data(self):
        """
        The view for deleting all Kapacitor and InfluxDB resources registered for a specific service function chain.

        :return: a JSON message to indicate whether deletion was successful
        """

        influx_host, influx_port = self.request.registry.settings['influx_host'], self.request.registry.settings['influx_port']
        clmc_service_port = self.request.registry.settings['clmc_service_port']

        # establish connection with InfluxDB
        client = InfluxDBClient(host=influx_host, port=influx_port)

        # fetch the URL parameters
        sfc_id = self.request.matchdict["sfc_id"]

        # delete alerts data
        alerts_delete_url = "http://localhost:{0}/alerts/{1}".format(clmc_service_port, sfc_id)
        log.info("Sending DELETE request to CLMC service for URL - {0}".format(alerts_delete_url))
        alerts_response = delete(alerts_delete_url)
        log.info("Received status code {0} from CLMC service for URL - {1}".format(alerts_response.status_code, alerts_delete_url))

        assert alerts_response.status_code == 200, "Alerts deletion returned status code different than 200 - {0}".format(alerts_response.text)

        # delete database and time-series data
        log.info("Deleting database {0}".format(sfc_id))
        client.drop_database(sfc_id)

        return {"msg": "Time-series data and alerts have been successfully deleted for SFC {0}.".format(sfc_id)}
